<?php

namespace Drupal\Tests\k8s\Functional\cloud\server_template;

use Drupal\cloud\Entity\CloudServerTemplateInterface;
use Drupal\Tests\cloud\Traits\CloudServerTemplateTrait;
use Drupal\Tests\k8s\Functional\K8sTestBase;

/**
 * Test K8s cloud launch templates (CloudServerTemplate).
 *
 * @group Cloud
 */
class CloudServerTemplateTest extends K8sTestBase {

  use CloudServerTemplateTrait;

  public const CLOUD_SERVER_TEMPLATES_REPEAT_COUNT = 2;

  private const CLOUD_SERVER_TEMPLATES_OPERATION_ADD = 'add';
  private const CLOUD_SERVER_TEMPLATES_OPERATION_EDIT = 'edit';

  /**
   * The generic login user.
   *
   * @var \Drupal\user\Entity\User
   */
  protected $webUser;

  /**
   * A user with administrative permissions.
   *
   * @var \Drupal\user\Entity\User
   */
  protected $adminUser;

  /**
   * {@inheritdoc}
   *
   * @throws \Exception
   */
  protected function getPermissions(): array {
    // Setup namespaces.
    $namespaces = $this->createNamespacesRandomTestFormData();
    $this->addNamespaceMockData($namespaces[0]);
    $this->createNamespaceTestEntity($namespaces[0]);
    $this->namespace = $namespaces[0]['name'];
    $this->getNamespaceMockData($namespaces[0]);

    return [

      // Cloud service provider.
      "view {$this->cloudContext}",
      'edit cloud service providers',

      // Namespace.
      "view k8s namespace {$this->namespace}",

      // Launch template.
      'add cloud server templates',
      'list cloud server template',
      'view any published cloud server templates',
      'view any unpublished cloud server templates',
      'launch cloud server template',
      'edit any cloud server templates',
      'delete any cloud server templates',
      'launch approved cloud server template',

      // Launch template revisions.
      'access cloud server template revisions',
      'revert all cloud server template revisions',
      'delete all cloud server template revisions',

      // Pod.
      'add k8s pod',
      'list k8s pod',
      'view any k8s pod',
      'edit any k8s pod',
      'delete any k8s pod',

      // Deployment.
      'add k8s deployment',
      'list k8s deployment',
      'view any k8s deployment',
      'edit any k8s deployment',
      'delete any k8s deployment',
    ];
  }

  /**
   * Set up test.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function setUp(): void {

    parent::setUp();

    $this->webUser = $this->createLoginUser();
    $this->adminUser = $this->createLoginUser(['approve launch k8s resources']);
    $this->drupalLogin($this->adminUser);

  }

  /**
   * CRUD test for k8s launch template.
   */
  public function testK8sServerTemplate(): void {
    $objects = $this->getK8sObjects();
    foreach ($objects ?: [] as $object) {
      try {
        $this->runServerTemplateCrudTest($object);
      }
      catch (\Exception $e) {
        throw new \RuntimeException($e->getMessage());
      }
    }
  }

  /**
   * CRUD test for k8s launch template launch.
   */
  public function testK8sServerTemplateLaunch(): void {
    $objects = $this->getK8sObjects();
    foreach ($objects ?: [] as $object) {
      try {
        $this->runServerTemplateLaunchTest($object);
      }
      catch (\Exception $e) {
        throw new \RuntimeException($e->getMessage());
      }
    }
  }

  /**
   * Workflow update test for k8s launch template launch.
   */
  public function testCloudServerTemplateWorkflowUpdate(): void {
    $objects = $this->getK8sObjects();
    $index = 0;
    foreach ($objects ?: [] as $index => $object) {
      try {
        $index++;
        $this->runServerTemplateLaunchWorkflowUpdateTest($object, $index);
      }
      catch (\Exception $e) {
        throw new \RuntimeException($e->getMessage());
      }
    }
  }

  /**
   * Delete resources test for k8s launch template launch.
   */
  public function testK8sServerTemplateDeleteResources(): void {
    for ($i = 0; $i < self::CLOUD_SERVER_TEMPLATES_REPEAT_COUNT; $i++) {
      try {
        $this->runServerTemplateDeleteResourcesTest();
      }
      catch (\Exception $e) {
        throw new \RuntimeException($e->getMessage());
      }
    }
  }

  /**
   * CRUD test for k8s pod launch template.
   *
   * @param string $object
   *   Object to test.
   *
   * @throws \Behat\Mink\Exception\ExpectationException
   * @throws \Behat\Mink\Exception\ResponseTextException
   */
  private function runServerTemplateCrudTest($object): void {

    $cloud_context = $this->cloudContext;

    // Create test.
    $pods = $this->createPodTestFormData(self::CLOUD_SERVER_TEMPLATES_REPEAT_COUNT, $this->namespace);
    $add = $this->createServerTemplateTestFormData($pods, 'yml', $object, [], self::CLOUD_SERVER_TEMPLATES_REPEAT_COUNT);
    for ($i = 0; $i < self::CLOUD_SERVER_TEMPLATES_REPEAT_COUNT; $i++) {
      // field_object is automatically calculated.
      unset($add[$i]['field_object']);
      $this->drupalGet("/clouds/design/server_template/$cloud_context/k8s/add");
      $this->submitForm(
        $add[$i],
        $this->t('Save'));

      $this->assertNoErrorMessage();
      $t_args = [
        '@type' => 'launch template',
        '%label' => $add[$i]['name[0][value]'],
      ];
      $this->assertSession()->pageTextContains(strip_tags($this->t('The @type %label has been created.', $t_args)));
    }

    // List test.
    $this->drupalGet("/clouds/design/server_template/$cloud_context");
    $this->assertNoErrorMessage();

    // Edit test.
    $pods = $this->createPodTestFormData(self::CLOUD_SERVER_TEMPLATES_REPEAT_COUNT, $this->namespace);
    $edit = $this->createServerTemplateTestFormData($pods, 'yml', $object, [], self::CLOUD_SERVER_TEMPLATES_REPEAT_COUNT);
    for ($i = 0; $i < self::CLOUD_SERVER_TEMPLATES_REPEAT_COUNT; $i++) {
      // field_object is automatically calculated.
      unset($edit[$i]['field_object']);

      // Go to listing page.
      $this->drupalGet("/clouds/design/server_template/$cloud_context");
      $this->clickLink('Edit', $i);

      $this->drupalGet($this->getUrl());
      $this->submitForm(
        $edit[$i],
        $this->t('Save'));

      $this->assertNoErrorMessage();
      $t_args = [
        '@type' => 'launch template',
        '%label' => $edit[$i]['name[0][value]'],
      ];
      $this->assertSession()->pageTextContains(strip_tags($this->t('The @type %label has been updated.', $t_args)));
    }

    // Delete test.
    for ($i = 0; $i < self::CLOUD_SERVER_TEMPLATES_REPEAT_COUNT; $i++) {

      // Go to listing page.
      $this->drupalGet("/clouds/design/server_template/$cloud_context");

      $this->clickLink('Delete', 0);
      $this->drupalGet($this->getUrl());
      $this->submitForm(
        [],
        $this->t('Delete'));
      $this->assertNoErrorMessage();
    }

    // Make sure the deleted templates are not listed anymore.
    $this->drupalGet("/clouds/design/server_template/$cloud_context");
    $this->assertNoErrorMessage();
    for ($j = 0; $j < self::CLOUD_SERVER_TEMPLATES_REPEAT_COUNT; $j++) {
      $this->assertSession()->pageTextNotContains($edit[$j]['name[0][value]']);
    }
  }

  /**
   * Launch test for k8s pod launch template.
   *
   * @param string $object
   *   Object to test.
   *
   * @throws \Behat\Mink\Exception\ExpectationException
   * @throws \Behat\Mink\Exception\ResponseTextException
   */
  private function runServerTemplateLaunchTest($object): void {
    $cloud_context = $this->cloudContext;

    // Create templates.
    switch ($object) {
      case 'pod':
        $data = $this->createPodTestFormData(self::CLOUD_SERVER_TEMPLATES_REPEAT_COUNT, $this->namespace);
        break;

      case 'deployment':
        $data = $this->createDeploymentTestFormData(self::CLOUD_SERVER_TEMPLATES_REPEAT_COUNT, $this->namespace);
        break;

      default:
        break;
    }

    $add = $this->createServerTemplateTestFormData($data, 'yml', $object, [], self::CLOUD_SERVER_TEMPLATES_REPEAT_COUNT);

    for ($i = 0; $i < self::CLOUD_SERVER_TEMPLATES_REPEAT_COUNT; $i++) {
      // field_object is automatically calculated.
      unset($add[$i]['field_object']);
      $this->drupalGet("/clouds/design/server_template/$cloud_context/k8s/add");
      $this->submitForm(
        $add[$i],
        $this->t('Save'));

      $this->assertNoErrorMessage();
      $t_args = [
        '@type' => 'launch template',
        '%label' => $add[$i]['name[0][value]'],
      ];
      $this->assertSession()->pageTextContains(strip_tags($this->t('The @type %label has been created.', $t_args)));
    }

    // Launch the templates.
    for ($i = 0; $i < self::CLOUD_SERVER_TEMPLATES_REPEAT_COUNT; $i++) {
      $this->drupalGet("/clouds/design/server_template/$cloud_context");
      // Update the mock data.
      switch ($object) {
        case 'pod':
          $this->getMetricsPodMockData([]);
          $this->addPodMockData($data[$i]);
          break;

        case 'deployment':
          $this->addDeploymentMockData($data[$i]);
          break;

        default;
          break;
      }

      // Navigate to the launch template, and launch it.
      $this->clickLink($data[$i]['name']);
      $this->clickLink('Launch');

      $this->drupalGet($this->getUrl());
      $this->submitForm(
        [],
        $this->t('Launch'));

      $this->assertNoErrorMessage();
      $t_args = [
        '@type' => ucfirst($object),
        '%label' => $add[$i]['name[0][value]'],
      ];
      $this->assertSession()->pageTextContains(strip_tags($this->t('The @type %label has been launched.', $t_args)));
    }
  }

  /**
   * Workflow update test for k8s pod launch template.
   *
   * @param string $object
   *   Object to test.
   * @param int $index
   *   The template index.
   *
   * @throws \Behat\Mink\Exception\ExpectationException
   * @throws \Behat\Mink\Exception\ResponseTextException
   */
  private function runServerTemplateLaunchWorkflowUpdateTest($object, $index): void {
    $cloud_context = $this->cloudContext;
    $repeat_count = 1;
    $module_name = 'k8s';

    // Create templates.
    switch ($object) {
      case 'pod':
        $data = $this->createPodTestFormData($repeat_count, $this->namespace);
        break;

      case 'deployment':
        $data = $this->createDeploymentTestFormData($repeat_count, $this->namespace);
        break;

      default:
        break;
    }

    $add = $this->createServerTemplateTestFormData($data, 'yml', $object, [], $repeat_count, CloudServerTemplateInterface::DRAFT);
    $edit = $add[0];

    $list_url = "/clouds/design/server_template/$cloud_context";
    $view_url = "/clouds/design/server_template/$cloud_context/$index";

    // 1.The generic user create a Launch Template.
    $this->drupalLogin($this->webUser);
    $this->updateLaunchTemplate($cloud_context, $edit, CloudServerTemplateInterface::CLOUD_SERVER_TEMPLATES_OPERATION_ADD, $module_name, $index);
    // Validate the status is Draft.
    $this->checkPageTextContains($list_url, $view_url, CloudServerTemplateInterface::DRAFT);

    // 2.The generic user change the Launch Template status
    // from Draft to Review.
    $this->drupalLogin($this->webUser);
    $edit['field_workflow_status'] = CloudServerTemplateInterface::REVIEW;
    $this->updateLaunchTemplate($cloud_context, $edit, CloudServerTemplateInterface::CLOUD_SERVER_TEMPLATES_OPERATION_EDIT, $module_name, $index);
    // Validate the status is Review.
    $this->checkPageTextContains($list_url, $view_url, CloudServerTemplateInterface::REVIEW);

    // 3.The generic user change the Launch Template status
    // from Review to Draft.
    $this->drupalLogin($this->webUser);
    $edit['field_workflow_status'] = CloudServerTemplateInterface::DRAFT;
    $this->updateLaunchTemplate($cloud_context, $edit, CloudServerTemplateInterface::CLOUD_SERVER_TEMPLATES_OPERATION_EDIT, $module_name, $index);
    // Validate the status is still Draft.
    $this->checkPageTextContains($list_url, $view_url, CloudServerTemplateInterface::DRAFT);

    // 4.The generic user change the Launch Template status
    // from Draft to Review.
    $this->drupalLogin($this->webUser);
    $edit['field_workflow_status'] = CloudServerTemplateInterface::REVIEW;
    $this->updateLaunchTemplate($cloud_context, $edit, CloudServerTemplateInterface::CLOUD_SERVER_TEMPLATES_OPERATION_EDIT, $module_name, $index);
    // Validate the status is still REVIEW.
    $this->checkPageTextContains($list_url, $view_url, CloudServerTemplateInterface::REVIEW);
    // Make sure that Generic user cannot Approve.
    $this->drupalGet($view_url);
    $this->assertSession()->linkNotExistsExact($this->t('Approve'));
    $this->assertSession()->linkByHrefNotExists("server_template/$cloud_context/$index/approve");

    // 5.The IT admin change the Launch Template status from Review to Approved.
    $this->drupalLogin($this->adminUser);
    // Make sure that the Admin user can Approve.
    $this->drupalGet($view_url);
    $this->assertSession()->linkExistsExact($this->t('Approve'));
    $this->assertSession()->linkByHrefExists("server_template/$cloud_context/$index/approve");
    $edit['field_workflow_status'] = CloudServerTemplateInterface::APPROVED;
    $this->updateLaunchTemplate($cloud_context, $edit, CloudServerTemplateInterface::CLOUD_SERVER_TEMPLATES_OPERATION_EDIT, $module_name, $index);
    // Validate the status is still APPROVED.
    $this->checkPageTextContains($list_url, $view_url, CloudServerTemplateInterface::APPROVED);
    // Validate Launch tab IS shown in the generic user's Launch Template.
    $this->drupalLogin($this->webUser);
    $this->drupalGet($view_url);
    $this->assertSession()->linkExistsExact($this->t('Launch'));
    $this->assertSession()->linkByHrefExists("server_template/$cloud_context/$index/launch");

    // 6.The generic user save the Launch Template as it is w/o any change.
    $this->drupalLogin($this->webUser);
    $this->updateLaunchTemplate($cloud_context, $edit, CloudServerTemplateInterface::CLOUD_SERVER_TEMPLATES_OPERATION_EDIT, $module_name, $index);
    // Validate the status is still APPROVED.
    $this->checkPageTextContains($list_url, $view_url, CloudServerTemplateInterface::APPROVED);

    // 7.The generic user save the Launch Template w/ some changes.
    $this->drupalLogin($this->webUser);
    $template_name = $edit['name[0][value]'];
    $edit['name[0][value]'] .= '-edit';
    $this->updateLaunchTemplate($cloud_context, $edit, CloudServerTemplateInterface::CLOUD_SERVER_TEMPLATES_OPERATION_EDIT, $module_name, $index);
    // Validate the status is still DRAFT.
    $this->checkPageTextContains($list_url, $view_url, CloudServerTemplateInterface::DRAFT);
    $edit['name[0][value]'] = $template_name;

    // 8.The generic user change the Launch Template status
    // from Draft to Review.
    $this->drupalLogin($this->webUser);
    $edit['field_workflow_status'] = CloudServerTemplateInterface::REVIEW;
    $this->updateLaunchTemplate($cloud_context, $edit, CloudServerTemplateInterface::CLOUD_SERVER_TEMPLATES_OPERATION_EDIT, $module_name, $index);
    // Validate the status is still REVIEW.
    $this->checkPageTextContains($list_url, $view_url, CloudServerTemplateInterface::REVIEW);
    // Make sure that Generic user cannot Approve.
    $this->drupalGet($view_url);
    $this->assertSession()->linkNotExistsExact($this->t('Approve'));
    $this->assertSession()->linkByHrefNotExists("server_template/$cloud_context/$index/approve");

    // 9.The IT admin change the Launch Template status from Review to Approved.
    $this->drupalLogin($this->adminUser);
    // Make sure that the Admin user can Approve.
    $this->drupalGet($view_url);
    $this->assertSession()->linkExistsExact($this->t('Approve'));
    $this->assertSession()->linkByHrefExists("server_template/$cloud_context/$index/approve");
    $edit['field_workflow_status'] = CloudServerTemplateInterface::APPROVED;
    $this->updateLaunchTemplate($cloud_context, $edit, CloudServerTemplateInterface::CLOUD_SERVER_TEMPLATES_OPERATION_EDIT, $module_name, $index);
    // Validate the status is still APPROVED.
    $this->checkPageTextContains($list_url, $view_url, CloudServerTemplateInterface::APPROVED);
    // Validate Launch tab IS shown in the generic user's Launch Template.
    $this->drupalLogin($this->webUser);
    $this->drupalGet($view_url);
    $this->assertSession()->linkExistsExact($this->t('Launch'));
    $this->assertSession()->linkByHrefExists("server_template/$cloud_context/$index/launch");

    // 10.The generic user launches an instance from Launch Template.
    $this->drupalLogin($this->webUser);
    $this->drupalGet("/clouds/design/server_template/$cloud_context/$index/launch");
    // Update the mock data.
    switch ($object) {
      case 'pod':
        $this->getMetricsPodMockData([]);
        $this->addPodMockData($data[0]);
        break;

      case 'deployment':
        $this->addDeploymentMockData($data[0]);
        break;

      default;
        break;
    }

    $this->submitForm(
      [],
      $this->t('Launch'));
    $this->assertNoErrorMessage();
    $t_args = [
      '@type' => ucfirst($object),
      '%label' => $edit['name[0][value]'],
    ];
    $this->assertSession()->pageTextContains(strip_tags($this->t('The @type %label has been launched.', $t_args)));
    // After launching the instance.
    $this->checkPageTextContains($list_url, $view_url, CloudServerTemplateInterface::DRAFT);

    // Delete template.
    $this->drupalGet("/clouds/design/server_template/$cloud_context/$index/delete");
    $this->submitForm(
      ['field_delete_option' => 'application'],
      $this->t('Delete'));
    $this->assertNoErrorMessage();
  }

  /**
   * Delete resources test for k8s pod launch template.
   *
   * @throws \Behat\Mink\Exception\ElementNotFoundException
   * @throws \Behat\Mink\Exception\ExpectationException
   * @throws \Behat\Mink\Exception\ResponseTextException
   */
  private function runServerTemplateDeleteResourcesTest(): void {
    $cloud_context = $this->cloudContext;
    $available_resources = \Drupal::service('k8s')->supportedCloudServerTemplates();

    // Because there is a special type "mixed", the number of resources
    // should be larger than 1.
    $keys = array_rand($available_resources, random_int(2, count($available_resources)));
    if (!is_array($keys)) {
      $keys = [$keys];
    }

    $keys = array_flip($keys);
    $objects = array_intersect_key($available_resources, $keys);

    // Create resources.
    $entities = [];
    foreach ($objects as $object => $name) {
      $count = random_int(1, self::CLOUD_SERVER_TEMPLATES_REPEAT_COUNT);
      $createFormDataFunction = "create{$name}TestFormData";
      if (!method_exists($this, $createFormDataFunction)) {
        continue;
      }

      $items = $this->$createFormDataFunction($count, $this->namespace);
      if ($object === 'pod') {
        $this->getMetricsPodMockData([]);
      }

      $addMockDataFunction = "add{$name}MockData";
      $createTestEntityFunction = "create{$name}TestEntity";
      $deleteMockDataFunction = "delete{$name}MockData";

      if (!method_exists($this, $addMockDataFunction)) {
        continue;
      }

      foreach ($items ?: [] as $item) {
        $this->$addMockDataFunction($item);
        $entities[] = $this->$createTestEntityFunction($item);
        $this->$deleteMockDataFunction($item);
      }
    }

    $adds = $this->createServerTemplateTestFormData([], 'git', $object, $entities);
    $adds[0]['name'] = $adds[0]['name[0][value]'];
    $adds[0]['cloud_context'] = $adds[0]['cloud_context[0][value]'];
    $server_template = $this->createServerTemplateTestEntity($adds[0]);
    $id = $server_template->id();
    $this->drupalGet("/clouds/design/server_template/$cloud_context/$id/delete");
    $this->assertNoErrorMessage();
    $this->assertSession()->pageTextContains(
      $this->t("Please make sure the following resources will be deleted if you select 'Delete both application and resources' or 'Delete only resources' option."));
    $this->assertSession()->fieldEnabled('field_delete_option');

    foreach ($entities ?: [] as $entity) {
      $this->assertSession()->pageTextContains(
        $this->t('@type: @name', [
          '@type' => $entity->getEntityType()->getSingularLabel(),
          '@name' => $entity->getName(),
          ':url' => $entity->toUrl('canonical')->toString(),
        ])
      );
    }
    $options = ['both', 'resources', 'application'];
    $option = $options[array_rand($options)];

    $this->drupalGet("/clouds/design/server_template/$cloud_context/$id/delete");
    $this->submitForm(
      ['field_delete_option' => $option],
      $this->t('Delete'));

    $this->assertNoErrorMessage();

    if ($option === 'both' || $option === 'resources') {
      foreach ($entities ?: [] as $entity) {
        $t_args = [
          '@type' => $entity->getEntityType()->getSingularLabel(),
          '@label' => $entity->label(),
          '@cloud_context' => $cloud_context,
        ];
        $this->assertSession()->pageTextContains(strip_tags($this->t('The @type @label on @cloud_context has been deleted.', $t_args)));
      }
    }

    if ($option === 'both' || $option === 'application') {

      $t_args = [
        '@type'  => $server_template->getEntityType()->getSingularLabel(),
        '@label' => $server_template->label(),
      ];
      $this->assertSession()->pageTextContains(strip_tags($this->t('The @type @label has been deleted.', $t_args)));
    }
  }

  /**
   * Get K8bs objects to test.
   *
   * @return array
   *   Array of K8s objects.
   */
  protected function getK8sObjects(): array {
    return [
      'pod',
      'deployment',
    ];
  }

}
